// 
//  MAC_MPEG2_AV - MPEG-2 hardware implementation for Xilinx multimedia board 
//  Copyright (C) 2007 McMaster University
// 
//==============================================================================
// 
// This file is part of MAC_MPEG2_AV
// 
// MAC_MPEG2_AV is distributed in the hope that it will be useful for further 
// research, but WITHOUT ANY WARRANTY; without even the implied warranty of 
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. MAC_MPEG2_AV is free; you 
// can redistribute it and/or modify it provided that proper reference is provided 
// to the authors. See the documents included in the "doc" folder for further details.
//
//==============================================================================

// adapted from the ISO implementation of an MPEG audio decoder
// see the README.txt file in the /Additional_Software for details

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifndef 				PI
#define         	PI				3.14159265358979
#endif

#define         	PI4			PI/4
#define         	PI64			PI/64

void AIFF_samples(FILE *outfile);
int  getbits(FILE *infile, int num_bits);
void Decode_Header(FILE *infile);
void Decode_Bitalloc_Scale(FILE *infile);
void Buffer_Deq_Denorm(FILE *infile, int Scale_Block);
int  SubBandSynthesis(int Synth_Counter);

int table_code[608] = {
	0x1, 0x0, 0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xB, 0xC, 0xD, 0x1, 0x0, 0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xB, 0xC, 0xD,
	0x1, 0x0, 0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xB, 0xC, 0xD, 0x1, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xD,
	0x1, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xD, 0x1, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xD,
	0x1, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xD, 0x1, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xD,
	0x1, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xD, 0x1, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xD,
	0x1, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xD, 0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD, 0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD,
	0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD, 0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD, 0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD, 0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD,
	0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD, 0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD, 0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD, 0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD,
	0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD, 0x2, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0xD, 0x3, 0x0, 0xE, 0xD, 0x3, 0x0, 0xE, 0xD, 0x3, 0x0, 0xE, 0xD, 0x3, 0x0, 0xE, 0xD,
	0x3, 0x0, 0xE, 0xD, 0x3, 0x0, 0xE, 0xD, 0x3, 0x0, 0xE, 0xD, 0x1, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xB, 0xC, 0x1, 0x0, 0xE, 0xF,
	0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xB, 0xC, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF,
	0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF,
	0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x1, 0x0, 0xE, 0x0,
	0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xB, 0x1, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xB, 0x1, 0x0, 0xE, 0x0,
	0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xB, 0x1, 0x0, 0xE, 0x0, 0xF, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xA, 0xB, 0x2, 0x0, 0xE, 0xF,
	0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF,
	0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x2, 0x0, 0xE, 0xF, 0x1, 0x2, 0x3, 0x4, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF,
	0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF,
	0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF, 0x3, 0x0, 0xE, 0xF };

int shift_counter;
int table_num;
int SB_Limit;
					
int bit_alloc[2][32];
int scale_index[2][3][32];
short fraction[2][3][32];
short samples[64];
					
int table_address(int index_i, int index_j)
{
	switch(table_num) {
		case 0 : 
			if (index_i < 11) return index_i * 16 + index_j;
			else if (index_i < 23) return (index_i - 11) * 8 + (index_j % 8) + 176;
			else return (index_i - 23) * 4 + (index_j % 4) + 272;
		case 1 : 
			if (index_i < 2) return index_i * 16 + index_j + 300;
			else return (index_i - 2) * 8 + (index_j % 8) + 332;
		case 2 : 
			if (index_i < 4) return index_i * 16 + index_j + 412;
			else if (index_i < 11) return (index_i - 4) * 8 + (index_j % 8) + 476;
			else return (index_i - 11) * 4 + (index_j % 4) + 532;
	}
	return 0;
}

int table_bits(int index_i, int index_j)
{
	if ((index_j / 2) == 0) { // bypass condition
		switch (table_code[table_address(index_i, index_j)]) {
			case 0 : return 5;
			case 1 : return 4;
			case 2 : return 3;
			case 3 : return 2;
		}
	} else {
		if (table_code[table_address(index_i, index_j)] == 15) return 10;
		else if (table_code[table_address(index_i, index_j)] == 14) return 7;
		else return table_code[table_address(index_i, index_j)] + 3;
	}
	return 0;
}

int table_steps_MSB(int index_i, int index_j)
{
	if ((index_j / 2) == 0) { // bypass condition
		switch (table_code[table_address(index_i, index_j)]) {
			case 0 : return 1;
		}
	} else {
		if (table_code[table_address(index_i, index_j)] == 15) return 3;
		else if (table_code[table_address(index_i, index_j)] == 14) return 2;
		else return table_code[table_address(index_i, index_j)] + 2;
	}
	return 0;
}

int table_steps(int index_i, int index_j)
{
	if ((index_j / 2) == 0) { // bypass condition
		switch (table_code[table_address(index_i, index_j)]) {
			case 0 : return 3;
		}
	} else {
		if (table_code[table_address(index_i, index_j)] == 15) return 9;
		else if (table_code[table_address(index_i, index_j)] == 14) return 5;
		else return (8 << table_code[table_address(index_i, index_j)]) - 1;
	}
	return 0;
}

int table_group(int index_i, int index_j)
{
	if ((index_j / 2) == 0) { // bypass condition
		switch (table_code[table_address(index_i, index_j)]) {
			case 0 : return 1;
		}
	} else {
		if (table_code[table_address(index_i, index_j)] == 15) return 1;
		else if (table_code[table_address(index_i, index_j)] == 14) return 1;
		else return 3;
	}
	return 0;
}

int table_quant(int index_i, int index_j)
{
	if ((index_j / 2) == 0) { // bypass condition
		switch (table_code[table_address(index_i, index_j)]) {
			case 0 : return 0;
		}
	} else {
		if (table_code[table_address(index_i, index_j)] == 15) return 3;
		else if (table_code[table_address(index_i, index_j)] == 14) return 1;
		else if (table_code[table_address(index_i, index_j)] == 0) return 2;
		else return table_code[table_address(index_i, index_j)] + 3;
	}
	return 0;
}

int main(int argc, char *argv[])
{
	int i, j;
	FILE *infile, *outfile;
	
	if (argc != 3) {
		printf("usage: mp2_software_FIX16 infile outfile \n");
		exit(0);
	}
		
	infile = fopen(argv[1], "rb");
	outfile = fopen(argv[2], "wb");
	if ((infile == NULL) || (outfile == NULL)) { 
		printf("Problem with file %s\n", argv[1]); exit(0); }
	shift_counter = 0; i = getbits(infile, 32);
	
	// skip first packet
	Decode_Header(infile); getbits(infile, 16);
	
	while (!feof(infile)) {
		Decode_Header(infile);
		Decode_Bitalloc_Scale(infile);
		for (i = 0; i < 12; i++) {
			Buffer_Deq_Denorm(infile, i);
			for (j = 0; j < 6; j++) {
				SubBandSynthesis(j);
				if (j % 2) AIFF_samples(outfile);
			}
		}
	}
	
	fclose(infile); fclose(outfile);
}

void AIFF_samples(FILE *outfile)
{
	int i;

	for (i = 0; i < 64; i++) 
		fprintf(outfile, "%c%c", 
			(samples[i] >> 8) & 0xFF, samples[i] & 0xFF);
}

int getbits(FILE *infile, int num_bits)
{
	static int file_buffer = 0;
	unsigned int temp;
	
	temp = file_buffer;
	temp >>= (32 - num_bits);
	file_buffer <<= num_bits;
	shift_counter += num_bits;
	while (shift_counter >= 8) {
		file_buffer = (((file_buffer >> shift_counter) << 8) | fgetc(infile)) 
			<< (shift_counter - 8);
		shift_counter -= 8;
	}
	
	return temp;
}	

void Decode_Header(FILE *infile)
{
	int i, j;
	int version;
	int layer;
	int error_protection;
	int bitrate_index, bitrate, brpch;
	int sfreq_code, sampling_frequency;
	int padding;
	int extension;
	int mode;
	int mode_extension;
	int copyright;
	int original;
	int emphasis;
	int stereo;
	
	int bitrate_table[2][3][15] = {{
		{0,32,48,56,64,80,96,112,128,144,160,176,192,224,256},
		{0,8,16,24,32,40,48,56,64,80,96,112,128,144,160},
		{0,8,16,24,32,40,48,56,64,80,96,112,128,144,160}},
		
		{{0,32,64,96,128,160,192,224,256,288,320,352,384,416,448},
		{0,32,48,56,64,80,96,112,128,160,192,224,256,320,384},
		{0,32,40,48,56,64,80,96,112,128,160,192,224,256,320}}};

	// remove sync
	while ((shift_counter % 8) != 0) i = getbits(infile, 1);
	i = 0; while (i == 0) {
		do {
			j = getbits(infile, 8);
		} while (j != 255);
		j = getbits(infile, 8);
		if ((j >> 4) == 15) i = 1;
	}

	printf("- Sync -\n");

	version = (j >> 3) & 0x1;						//version
	layer = (j >> 1) &0x3; 							// layer
	error_protection = j & 0x1;					// error_protection
	bitrate_index = getbits(infile, 4); 		// bitrate index
	sfreq_code = getbits(infile, 2);				// sampling frequency
	padding = getbits(infile, 1); 				// padding
	extension = getbits(infile, 1); 				// extension
	mode = getbits(infile, 2); 					// mode
	mode_extension = getbits(infile, 2); 		// mode extension
	copyright = getbits(infile, 1); 				// copyright
	original = getbits(infile, 1); 				// original
	emphasis = getbits(infile, 2); 				// emphasis
	
	if (version == 1) {
		switch(sfreq_code) {
			case 0 : sampling_frequency = 44100;
			case 1 : sampling_frequency = 48000;
			case 2 : sampling_frequency = 32000;
			case 3 : sampling_frequency = 00000;
		}
	} else {
		switch(sfreq_code) {
			case 0 : sampling_frequency = 22050;
			case 1 : sampling_frequency = 24000;
			case 2 : sampling_frequency = 16000;
			case 3 : sampling_frequency = 00000;
		}
	}
	
  	stereo = (mode == 3) ? 1 : 2;
	bitrate = bitrate_table[version][layer-1][bitrate_index];
	brpch = bitrate / stereo;
	
	if (version == 1) {
		if (
			((sfreq_code == 1) && (brpch >= 56)) || 
			((brpch >= 56) && (brpch <= 80))
		) table_num = 0;
		else if ((sfreq_code != 1) && (brpch >= 96)) table_num = 1;
		else if ((sfreq_code != 2) && (brpch >= 48)) table_num = 2;
		else table_num = 3;
	} else table_num = 4;
	
	switch(table_num) {
		case 0 : SB_Limit = 27; break;
		case 1 : SB_Limit = 30; break;
		case 2 : SB_Limit = 8; break;
		case 3 : SB_Limit = 12; break;
		case 4 : SB_Limit = 30; break;
	}
};

void Decode_Bitalloc_Scale(FILE *infile)
{
	int i, j;
	int scfsi[2][32];
	
	// Decode bitalloc
	for (i = 0; i < 32; i++) 
		for (j = 0; j < 2; j++)
			if (i < SB_Limit)
				bit_alloc[j][i] = getbits(infile, table_bits(i, 0));
			else bit_alloc[j][i] = 0;

	// Decode scfsi
	for (i = 0; i < 32; i++) 
		for (j = 0; j < 2; j++)
			if ((i < SB_Limit) && (bit_alloc[j][i])) 
				scfsi[j][i] = getbits(infile, 2);
			else scfsi[j][i] = 0;

	// Decode scale_index
	for (i = 0; i < 32; i++) 
		for (j = 0; j < 2; j++)
			if ((i < SB_Limit) && (bit_alloc[j][i])) {
				switch (scfsi[j][i]) {
					case 0 : 
						scale_index[j][0][i] = getbits(infile, 6);
             		scale_index[j][1][i] = getbits(infile, 6);
             		scale_index[j][2][i] = getbits(infile, 6);
             		break;
             	case 1 : 
						scale_index[j][0][i] =
						scale_index[j][1][i] = getbits(infile, 6);
						scale_index[j][2][i] = getbits(infile, 6);
						break;
					case 3 : 
						scale_index[j][0][i] = getbits(infile, 6);
             		scale_index[j][1][i] =
						scale_index[j][2][i] = getbits(infile, 6);
						break;
					case 2 : 
						scale_index[j][0][i] =
              		scale_index[j][1][i] =
                  scale_index[j][2][i] = getbits(infile, 6);
						break;
				}
			} else {
				scale_index[j][0][i] =
        		scale_index[j][1][i] =
            scale_index[j][2][i] = 63;
			}
}

void Buffer_Deq_Denorm(FILE *infile, int Scale_Block)
{
	double c_double[17] = { 
		1.33333333333, 1.60000000000, 1.14285714286,
		1.77777777777, 1.06666666666, 1.03225806452,
		1.01587301587, 1.00787401575, 1.00392156863,
		1.00195694716, 1.00097751711, 1.00048851979,
		1.00024420024, 1.00012208522, 1.00006103888,
		1.00003051851, 1.00001525902 };

	double d_double[17] = { 
		0.500000000,   0.500000000,   0.250000000, 
		0.500000000,   0.125000000,   0.062500000, 
		0.031250000,   0.015625000,   0.007812500, 
		0.003906250,   0.001953125,   0.0009765625,
		0.00048828125, 0.00024414063, 0.00012207031,
		0.00006103516, 0.00003051758 };

	double multiple_double[64] = { 
		2.00000000000000, 1.58740105196820, 1.25992104989487, 1.00000000000000, 
		0.79370052598410, 0.62996052494744, 0.50000000000000, 0.39685026299205, 
		0.31498026247372, 0.25000000000000, 0.19842513149602, 0.15749013123686, 
		0.12500000000000, 0.09921256574801, 0.07874506561843, 0.06250000000000, 
		0.04960628287401, 0.03937253280921, 0.03125000000000, 0.02480314143700, 
		0.01968626640461, 0.01562500000000, 0.01240157071850, 0.00984313320230, 
		0.00781250000000, 0.00620078535925, 0.00492156660115, 0.00390625000000, 
		0.00310039267963, 0.00246078330058, 0.00195312500000, 0.00155019633981, 
		0.00123039165029, 0.00097656250000, 0.00077509816991, 0.00061519582514, 
		0.00048828125000, 0.00038754908495, 0.00030759791257, 0.00024414062500, 
		0.00019377454248, 0.00015379895629, 0.00012207031250, 0.00009688727124, 
		0.00007689947814, 0.00006103515625, 0.00004844363562, 0.00003844973907, 
		0.00003051757813, 0.00002422181781, 0.00001922486954, 0.00001525878906, 
		0.00001211090890, 0.00000961243477, 0.00000762939453, 0.00000605545445, 
		0.00000480621738, 0.00000381469727, 0.00000302772723, 0.00000240310869, 
		0.00000190734863, 0.00000151386361, 0.00000120155435, 1E-20 };

	static int init_flag = 1;
	static unsigned short int c[17]; 
	static unsigned short int d[17];
	static unsigned short int multiple[64];

	int i, j, m, k, x, c2, nlevels;
	int sample, temp1;
	unsigned int temp2;

	if (init_flag) {
		for (i = 0; i < 17; i++) {
			c[i] = (unsigned short int)((c_double[i]-1.0) * (double)(1 << 16));
			d[i] = (unsigned short int)(d_double[i] * (double)(1 << 15));
		}
		for (i = 0; i < 64; i++) {
			temp2 = (unsigned int)(multiple_double[i] * (double)(1 << 14));
			if (temp2 >= 65536) 	multiple[i] = 65535;
			else 						multiple[i] = (unsigned short int)temp2;
		}
		init_flag = 0;
	}

	for (i = 0; i < 32; i++) 
		for (j = 0; j < 2; j++)
			for (m = 0; m < 3; m++)
				if ((i < SB_Limit) && (bit_alloc[j][i])) {
					if (table_group(i, bit_alloc[j][i]) == 3) {
						k = table_bits(i, bit_alloc[j][i]);

						// buffer sample						
						sample = getbits(infile, k);
					} else {
						nlevels = table_steps(i, bit_alloc[j][i]);
						k = table_bits(i, bit_alloc[j][i]);
						if (m == 0) c2 = getbits(infile, k);

						// buffer sample						
						sample = c2 % nlevels;
						c2 /= nlevels;
					}

					// dequantise sample	
					x = table_steps_MSB(i, bit_alloc[j][i]);
					sample = ((sample << (31 - x)) ^ (1 << 31)) >> 16;
					sample += d[table_quant(i, bit_alloc[j][i])];
					temp1 = (int)((unsigned int)c[table_quant(i, bit_alloc[j][i])] + (1L << 16)) * 
						(unsigned int)multiple[scale_index[j][Scale_Block >> 2][i]];
					
					// denormalize sample
					temp1 = (temp1 + (1 << 14)) >> 15;
					temp1 *= sample;
					temp1 = (temp1 + (1 << 14)) >> 15;

					if (temp1 >= 32768) 			fraction[j][m][i] = 32767;
					else if (temp1 < -32768)	fraction[j][m][i] = -32768;
					else 								fraction[j][m][i] = (short int)temp1;
			} else fraction[j][m][i] = 0;
}      

int SubBandSynthesis(int Synth_Counter)
{
	static int bufOffset[2] = {64,64};
	static short int buf[2][1024];
	short int *bufOffsetPtr;
	
	static double filter_double[64][32]; 
	double window_double[512] = {
		 0.000000000, -0.000015259, -0.000015259, -0.000015259, -0.000015259, -0.000015259, -0.000015259, -0.000030518, 
		-0.000030518, -0.000030518, -0.000030518, -0.000045776, -0.000045776, -0.000061035, -0.000061035, -0.000076294, 
		-0.000076294, -0.000091553, -0.000106812, -0.000106812, -0.000122070, -0.000137329, -0.000152588, -0.000167847, 
		-0.000198364, -0.000213623, -0.000244141, -0.000259399, -0.000289917, -0.000320435, -0.000366211, -0.000396729, 
		-0.000442505, -0.000473022, -0.000534058, -0.000579834, -0.000625610, -0.000686646, -0.000747681, -0.000808716, 
		-0.000885010, -0.000961304, -0.001037598, -0.001113892, -0.001205444, -0.001296997, -0.001388550, -0.001480103, 
		-0.001586914, -0.001693726, -0.001785278, -0.001907349, -0.002014160, -0.002120972, -0.002243042, -0.002349854, 
		-0.002456665, -0.002578735, -0.002685547, -0.002792358, -0.002899170, -0.002990723, -0.003082275, -0.003173828, 
		 0.003250122,  0.003326416,  0.003387451,  0.003433228,  0.003463745,  0.003479004,  0.003479004,  0.003463745, 
		 0.003417969,  0.003372192,  0.003280640,  0.003173828,  0.003051758,  0.002883911,  0.002700806,  0.002487183, 
		 0.002227783,  0.001937866,  0.001617432,  0.001266479,  0.000869751,  0.000442505, -0.000030518, -0.000549316, 
		-0.001098633, -0.001693726, -0.002334595, -0.003005981, -0.003723145, -0.004486084, -0.005294800, -0.006118774, 
		-0.007003784, -0.007919312, -0.008865356, -0.009841919, -0.010848999, -0.011886597, -0.012939453, -0.014022827, 
		-0.015121460, -0.016235352, -0.017349243, -0.018463135, -0.019577026, -0.020690918, -0.021789551, -0.022857666, 
		-0.023910522, -0.024932861, -0.025909424, -0.026840210, -0.027725220, -0.028533936, -0.029281616, -0.029937744, 
		-0.030532837, -0.031005859, -0.031387329, -0.031661987, -0.031814575, -0.031845093, -0.031738281, -0.031478882, 
		 0.031082153,  0.030517578,  0.029785156,  0.028884888,  0.027801514,  0.026535034,  0.025085449,  0.023422241, 
		 0.021575928,  0.019531250,  0.017257690,  0.014801025,  0.012115479,  0.009231567,  0.006134033,  0.002822876, 
		-0.000686646, -0.004394531, -0.008316040, -0.012420654, -0.016708374, -0.021179199, -0.025817871, -0.030609131, 
		-0.035552979, -0.040634155, -0.045837402, -0.051132202, -0.056533813, -0.061996460, -0.067520142, -0.073059082, 
		-0.078628540, -0.084182739, -0.089706421, -0.095169067, -0.100540161, -0.105819702, -0.110946655, -0.115921021, 
		-0.120697021, -0.125259399, -0.129562378, -0.133590698, -0.137298584, -0.140670776, -0.143676758, -0.146255493, 
		-0.148422241, -0.150115967, -0.151306152, -0.151962280, -0.152069092, -0.151596069, -0.150497437, -0.148773193, 
		-0.146362305, -0.143264771, -0.139450073, -0.134887695, -0.129577637, -0.123474121, -0.116577148, -0.108856201, 
		 0.100311279,  0.090927124,  0.080688477,  0.069595337,  0.057617187,  0.044784546,  0.031082153,  0.016510010, 
		 0.001068115, -0.015228271, -0.032379150, -0.050354004, -0.069168091, -0.088775635, -0.109161377, -0.130310059, 
		-0.152206421, -0.174789429, -0.198059082, -0.221984863, -0.246505737, -0.271591187, -0.297210693, -0.323318481, 
		-0.349868774, -0.376800537, -0.404083252, -0.431655884, -0.459472656, -0.487472534, -0.515609741, -0.543823242, 
		-0.572036743, -0.600219727, -0.628295898, -0.656219482, -0.683914185, -0.711318970, -0.738372803, -0.765029907, 
		-0.791213989, -0.816864014, -0.841949463, -0.866363525, -0.890090942, -0.913055420, -0.935195923, -0.956481934, 
		-0.976852417, -0.996246338, -1.014617920, -1.031936646, -1.048156738, -1.063217163, -1.077117920, -1.089782715, 
		-1.101211548, -1.111373901, -1.120223999, -1.127746582, -1.133926392, -1.138763428, -1.142211914, -1.144287109, 
		 1.144989014,  1.144287109,  1.142211914,  1.138763428,  1.133926392,  1.127746582,  1.120223999,  1.111373901, 
		 1.101211548,  1.089782715,  1.077117920,  1.063217163,  1.048156738,  1.031936646,  1.014617920,  0.996246338, 
		 0.976852417,  0.956481934,  0.935195923,  0.913055420,  0.890090942,  0.866363525,  0.841949463,  0.816864014, 
		 0.791213989,  0.765029907,  0.738372803,  0.711318970,  0.683914185,  0.656219482,  0.628295898,  0.600219727, 
		 0.572036743,  0.543823242,  0.515609741,  0.487472534,  0.459472656,  0.431655884,  0.404083252,  0.376800537, 
		 0.349868774,  0.323318481,  0.297210693,  0.271591187,  0.246505737,  0.221984863,  0.198059082,  0.174789429, 
		 0.152206421,  0.130310059,  0.109161377,  0.088775635,  0.069168091,  0.050354004,  0.032379150,  0.015228271, 
		-0.001068115, -0.016510010, -0.031082153, -0.044784546, -0.057617187, -0.069595337, -0.080688477, -0.090927124, 
		 0.100311279,  0.108856201,  0.116577148,  0.123474121,  0.129577637,  0.134887695,  0.139450073,  0.143264771, 
		 0.146362305,  0.148773193,  0.150497437,  0.151596069,  0.152069092,  0.151962280,  0.151306152,  0.150115967, 
		 0.148422241,  0.146255493,  0.143676758,  0.140670776,  0.137298584,  0.133590698,  0.129562378,  0.125259399, 
		 0.120697021,  0.115921021,  0.110946655,  0.105819702,  0.100540161,  0.095169067,  0.089706421,  0.084182739, 
		 0.078628540,  0.073059082,  0.067520142,  0.061996460,  0.056533813,  0.051132202,  0.045837402,  0.040634155, 
		 0.035552979,  0.030609131,  0.025817871,  0.021179199,  0.016708374,  0.012420654,  0.008316040,  0.004394531, 
		 0.000686646, -0.002822876, -0.006134033, -0.009231567, -0.012115479, -0.014801025, -0.017257690, -0.019531250, 
		-0.021575928, -0.023422241, -0.025085449, -0.026535034, -0.027801514, -0.028884888, -0.029785156, -0.030517578, 
		 0.031082153,  0.031478882,  0.031738281,  0.031845093,  0.031814575,  0.031661987,  0.031387329,  0.031005859, 
		 0.030532837,  0.029937744,  0.029281616,  0.028533936,  0.027725220,  0.026840210,  0.025909424,  0.024932861, 
		 0.023910522,  0.022857666,  0.021789551,  0.020690918,  0.019577026,  0.018463135,  0.017349243,  0.016235352, 
		 0.015121460,  0.014022827,  0.012939453,  0.011886597,  0.010848999,  0.009841919,  0.008865356,  0.007919312, 
		 0.007003784,  0.006118774,  0.005294800,  0.004486084,  0.003723145,  0.003005981,  0.002334595,  0.001693726, 
		 0.001098633,  0.000549316,  0.000030518, -0.000442505, -0.000869751, -0.001266479, -0.001617432, -0.001937866, 
		-0.002227783, -0.002487183, -0.002700806, -0.002883911, -0.003051758, -0.003173828, -0.003280640, -0.003372192, 
		-0.003417969, -0.003463745, -0.003479004, -0.003479004, -0.003463745, -0.003433228, -0.003387451, -0.003326416, 
		 0.003250122,  0.003173828,  0.003082275,  0.002990723,  0.002899170,  0.002792358,  0.002685547,  0.002578735, 
		 0.002456665,  0.002349854,  0.002243042,  0.002120972,  0.002014160,  0.001907349,  0.001785278,  0.001693726, 
		 0.001586914,  0.001480103,  0.001388550,  0.001296997,  0.001205444,  0.001113892,  0.001037598,  0.000961304, 
		 0.000885010,  0.000808716,  0.000747681,  0.000686646,  0.000625610,  0.000579834,  0.000534058,  0.000473022, 
		 0.000442505,  0.000396729,  0.000366211,  0.000320435,  0.000289917,  0.000259399,  0.000244141,  0.000213623, 
		 0.000198364,  0.000167847,  0.000152588,  0.000137329,  0.000122070,  0.000106812,  0.000106812,  0.000091553, 
		 0.000076294,  0.000076294,  0.000061035,  0.000061035,  0.000045776,  0.000045776,  0.000030518,  0.000030518, 
		 0.000030518,  0.000030518,  0.000015259,  0.000015259,  0.000015259,  0.000015259,  0.000015259,  0.000015259 };
	int i, j, k, sum, clip;
	static int init_flag = 1;
	static short int window[512];
	static unsigned short int filter[32];
	int filter_address;

	if (init_flag) {
		for (i=0; i<64; i++)
			for (k=0; k<32; k++) {
				if ((filter_double[i][k] = 1e9*cos((double)((PI64*i+PI4)*(2*k+1)))) >= 0)
					modf(filter_double[i][k]+0.5, &filter_double[i][k]);
				else
					modf(filter_double[i][k]-0.5, &filter_double[i][k]);
				filter_double[i][k] *= 1e-9;
			}

		for (i = 0; i < 32; i++) {
			sum = (int)(cos((double)i * 3.14159265 / 64.0) * (double)(1 << 16));
			if (sum >= 65536) filter[i] = 65535;
			else filter[i] = (unsigned short int)sum;
		}

		for (i = 0; i < 512; i++)
			window[i] = (short int)(window_double[i] * (double)(1 << 14));

      for (i = 0; i < 1024; i++) buf[0][i] = buf[1][i] = 0;
		
		init_flag = 0;
	}

	bufOffset[Synth_Counter % 2] = (bufOffset[Synth_Counter % 2] - 64) & 0x3ff;
	bufOffsetPtr = &(buf[Synth_Counter % 2][bufOffset[Synth_Counter % 2]]);

	for (i = 0; i < 64; i++) {
		sum = 0; filter_address = i + 16;
		for (k = 0; k < 32; k++) {
			if (((filter_address & 0x7F) != 32) && ((filter_address & 0x7F) != 96)) {
				j = ((filter_address >> 5) & 1) ? 
					-(filter_address & 0x1F) : (filter_address & 0x1F);
				if (((filter_address >> 5) & 1) ^ ((filter_address >> 6) & 1)) {
					sum -= ((int)fraction[Synth_Counter%2][Synth_Counter/2][k] * (int)(filter[j & 0x1F])) >> 4;
				} else {
					sum += ((int)fraction[Synth_Counter%2][Synth_Counter/2][k] * (int)(filter[j & 0x1F])) >> 4;
				} 				
			}
			filter_address = filter_address + 2*(i + 16);
		}
		sum = (sum + (1 << 11)) >> 12;

		if (sum >= 32768) 		bufOffsetPtr[i] = 32767;
		else if (sum < -32768) 	bufOffsetPtr[i] = -32768;
		else 							bufOffsetPtr[i] = (short)sum;
	}

	for (j = 0; j < 32; j++) {
		sum = 0;
		for (i = 0; i < 16; i++) {
			k = j + (i << 5);
			sum += ((window[k] << 1) * buf[Synth_Counter % 2][
				( (k + ( ((i+1)>>1) <<6) ) + bufOffset[Synth_Counter % 2]) & 0x3ff]) >> 4;
		}
		sum = (sum + (1 << 10)) >> 11;

		if (sum >= 32768)      {samples[(Synth_Counter % 2) + 2*j] =  32767; ++clip;}
		else if (sum < -32768) {samples[(Synth_Counter % 2) + 2*j] = -32768; ++clip;}
		else                    samples[(Synth_Counter % 2) + 2*j] =(short)sum;
	}
	return(clip);
}
