// 
//  MAC_MPEG2_AV - MPEG-2 hardware implementation for Xilinx multimedia board 
//  Copyright (C) 2007 McMaster University
// 
//==============================================================================
// 
// This file is part of MAC_MPEG2_AV
// 
// MAC_MPEG2_AV is distributed in the hope that it will be useful for further 
// research, but WITHOUT ANY WARRANTY; without even the implied warranty of 
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. MAC_MPEG2_AV is free; you 
// can redistribute it and/or modify it provided that proper reference is provided 
// to the authors. See the documents included in the "doc" folder for further details.
//
//==============================================================================

// backend.c - upsampling and colourspace conversions from YUV to RGB
// Adapted from the MPEG decoding software of the MPEG Software
// Simulation Group - see the README.txt file in the /Additional_Software
// folder for more info.

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#define GLOBAL

/* define NON_ANSI_COMPILER for compilers without function prototyping */
/* #define NON_ANSI_COMPILER */

#ifdef NON_ANSI_COMPILER
#define _ANSI_ARGS_(x) ()
#else
#define _ANSI_ARGS_(x) x
#endif

#define RB "rb"
#define WB "wb"

#ifndef O_BINARY
#define O_BINARY 0
#endif

#define CHROMA420 1
#define CHROMA422 2
#define CHROMA444 3
#define FILENAME_LENGTH 256

#ifndef GLOBAL
#define EXTERN extern
#else
#define EXTERN
#endif

EXTERN char Output_Picture_Filename[FILENAME_LENGTH];
EXTERN int Coded_Picture_Width;
EXTERN int Coded_Picture_Height;
EXTERN int horizontal_size;
EXTERN int vertical_size;
EXTERN int progressive_sequence;
EXTERN int chroma_format;
EXTERN int progressive_frame;
EXTERN int Frame_Store_Flag;
EXTERN unsigned char *src_global[3];
EXTERN char Error_Text[256];
EXTERN unsigned char *Clip;
EXTERN int matrix_coefficients;
EXTERN int MPEG2_Flag;

EXTERN int Inverse_Table_6_9[8][4]
#ifdef GLOBAL
=
{
  {117504, 138453, 13954, 34903}, /* no sequence_display_extension */
  {117504, 138453, 13954, 34903}, /* ITU-R Rec. 709 (1990) */
  {104597, 132201, 25675, 53279}, /* unspecified */
  {104597, 132201, 25675, 53279}, /* reserved */
  {104448, 132798, 24759, 53109}, /* FCC */
  {104597, 132201, 25675, 53279}, /* ITU-R Rec. 624-4 System B, G */
  {104597, 132201, 25675, 53279}, /* SMPTE 170M */
  {117579, 136230, 16907, 35559}  /* SMPTE 240M (1987) */
}
#endif
;

/* private prototypes */
static void Write_Frame _ANSI_ARGS_((unsigned char *src[], int frame));
static void store_ppm _ANSI_ARGS_((char *outname, unsigned char *src[],
  int offset, int incr, int height));
static unsigned char getbyte _ANSI_ARGS_((void));
static void putbyte _ANSI_ARGS_((int c));
static void putword _ANSI_ARGS_((int w));
static void conv422to444 _ANSI_ARGS_((unsigned char *src, unsigned char *dst));
static void conv420to422 _ANSI_ARGS_((unsigned char *src, unsigned char *dst));
void Error _ANSI_ARGS_((char *text));

#define OBFRSIZE 4096
static unsigned char obfr[OBFRSIZE];
static unsigned char *optr;
static int outfile, infile;

int main(int argc, char *argv[])
{
  int i, j, k, size, tempint;
  int layout, hsize, component, height;
  unsigned char *p;
  char tmpname[FILENAME_LENGTH], tempstr[100];
  FILE *infofile;
  
  if (argc != 2) {
    printf("usage: backend filename\n");
    exit(0);
  }
  
  /* Clip table */
  if (!(Clip=(unsigned char *)malloc(1024))) Error("Clip[] malloc failed\n");
  Clip += 384;
  for (i=-384; i<640; i++) Clip[i] = (i<0) ? 0 : ((i>255) ? 255 : i);

  // open information file
  sprintf(tmpname, "%s.info", argv[1]);
  if ((infofile = fopen(tmpname, "rt")) == NULL) {
     sprintf(Error_Text,"info file %s not found\n", tmpname);
     Error(Error_Text);
  }
  fscanf(infofile, "%s", tempstr); fscanf(infofile, "%s", tempstr);
  
  fscanf(infofile, "%s", Output_Picture_Filename);
  fscanf(infofile, "%s", tempstr); fscanf(infofile, "%s", tempstr);

  fscanf(infofile, "%d", &matrix_coefficients);
  fscanf(infofile, "%s", tempstr); fscanf(infofile, "%s", tempstr);

  fscanf(infofile, "%d", &MPEG2_Flag);
  fscanf(infofile, "%s", tempstr);

  fscanf(infofile, "%d", &layout);
  fscanf(infofile, "%s", tempstr);

  fscanf(infofile, "%d", &progressive_sequence);
  fscanf(infofile, "%d", &progressive_frame);
  fscanf(infofile, "%d", &Frame_Store_Flag);
  fscanf(infofile, "%s", tempstr); fscanf(infofile, "%s", tempstr); fscanf(infofile, "%s", tempstr);

  fscanf(infofile, "%d", &Coded_Picture_Width); 
  fscanf(infofile, "%s", tempstr);
  fscanf(infofile, "%d", &Coded_Picture_Height);
  fscanf(infofile, "%s", tempstr);

  fscanf(infofile, "%d", &horizontal_size);
  fscanf(infofile, "%s", tempstr); fscanf(infofile, "%s", tempstr);
  fscanf(infofile, "%d", &vertical_size);
  fscanf(infofile, "%s", tempstr);

  fscanf(infofile, "%d", &chroma_format);
  fclose(infofile);    

  printf("Output filename: %s\n", Output_Picture_Filename);
  printf("Matrix Coefficients: %d\n", matrix_coefficients);
  printf("MPEG2 flag: %d\n", MPEG2_Flag);
  printf("Layout: %d\n", layout);
  printf("Flags: %d %d %d\n", progressive_sequence, progressive_frame, Frame_Store_Flag);
  printf("Coded picture size: %d x %d\n", Coded_Picture_Width, Coded_Picture_Height);
  printf("Horizontal %d x Vertical %d\n", horizontal_size, vertical_size);
  printf("Format: %d = ", chroma_format);
  switch (chroma_format) {
    case (CHROMA444) : printf("444 Format\n"); break;
    case (CHROMA422) : printf("422 Format\n"); break;
    case (CHROMA420) : printf("420 Format\n"); break;
    default : 
      printf("Unrecognized format\n");
      break;
  }

  for (i=0; i<3; i++) {
    size = (i == 0) ? (Coded_Picture_Width*Coded_Picture_Height) : 
      (((chroma_format==CHROMA444) ? Coded_Picture_Width : Coded_Picture_Width>>1) * 
       ((chroma_format!=CHROMA420) ? Coded_Picture_Height : Coded_Picture_Height>>1));
    if (!(src_global[i] = (unsigned char *)malloc(size)))
      Error("src[] malloc failed\n");
  }

  // open memory file
  if (layout == 3) sprintf(tmpname, "%s.YUV3", argv[1]);
  else if (layout == 4) sprintf(tmpname, "%s.YUV4", argv[1]);

  if ((infile = open(tmpname,O_RDONLY|O_BINARY))<0) {
     sprintf(Error_Text,"info file %s not found\n", tmpname);
     Error(Error_Text);
  }

  k = 0; hsize = horizontal_size; height = vertical_size;
  for (component = 0; component < 3; component++) {
    p = src_global[component];
    for (i = 0; i < height; i++)
      for (j = 0; j < hsize; j++) {
        if ((layout == 3) && (k%4 == 0)) { 
          tempint = (int) getbyte(); k++;
        }
        *p++ = getbyte();
        k++;
      }
    while (k%4) {
      tempint = (int) getbyte(); k++;
    }
    if (component == 0) {
      if (chroma_format!=CHROMA444) hsize>>=1;
      if (chroma_format==CHROMA420) height>>=1;
    }
  }

  close(infile);   
/*
for (k = 0; k < 3; k++) {
   printf("k = %d\n", k);
   p = src_global[k];
   for (i = 0; i < ((k == 0) ? vertical_size : vertical_size >> 1); i++)
      for (j = 0; j < ((k == 0) ? horizontal_size : horizontal_size >> 1); j++)
         printf("i,j = %d,%d: %x\n", i, j, *p++);
}
*/
  // Write_Frame
  Write_Frame(src_global,0);

  for (i=0; i<3; i++) free(src_global[i]);
  
  return 0;   
}

/*
 * store a picture as either one frame or two fields
 */
void Write_Frame(src,frame)
unsigned char *src[];
int frame;
{
  char outname[FILENAME_LENGTH];

  if (progressive_sequence || progressive_frame || Frame_Store_Flag)
  {
    /* progressive */
    sprintf(outname,Output_Picture_Filename,frame,'f');
    store_ppm(outname,src,0,Coded_Picture_Width,vertical_size);
  }
  else
  {
    /* interlaced */
    sprintf(outname,Output_Picture_Filename,frame,'a');
    store_ppm(outname,src,0,Coded_Picture_Width<<1,vertical_size>>1);

    sprintf(outname,Output_Picture_Filename,frame,'b');
    store_ppm(outname,src,
      Coded_Picture_Width,Coded_Picture_Width<<1,vertical_size>>1);
  }
}

/*
 * store as PPM (PBMPLUS)
 */
static void store_ppm(outname,src,offset,incr,height)
char *outname;
unsigned char *src[];
int offset, incr, height;
{
  int i, j;
  int y, u, v, r, g, b;
  int crv, cbu, cgu, cgv;
  unsigned char *py, *pu, *pv;
  static unsigned char tga24[14] = {0,0,2,0,0,0,0, 0,0,0,0,0,24,32};
  char header[FILENAME_LENGTH];
  static unsigned char *u422, *v422, *u444, *v444;

  if (chroma_format==CHROMA444)
  {
    u444 = src[1];
    v444 = src[2];
  }
  else
  {
    if (!u444)
    {
      if (chroma_format==CHROMA420)
      {
        if (!(u422 = (unsigned char *)malloc((Coded_Picture_Width>>1)
                                             *Coded_Picture_Height)))
          Error("malloc failed");
        if (!(v422 = (unsigned char *)malloc((Coded_Picture_Width>>1)
                                             *Coded_Picture_Height)))
          Error("malloc failed");
      }

      if (!(u444 = (unsigned char *)malloc(Coded_Picture_Width
                                           *Coded_Picture_Height)))
        Error("malloc failed");

      if (!(v444 = (unsigned char *)malloc(Coded_Picture_Width
                                           *Coded_Picture_Height)))
        Error("malloc failed");
    }

    if (chroma_format==CHROMA420)
    {
      conv420to422(src[1],u422);
      conv420to422(src[2],v422);
      conv422to444(u422,u444);
      conv422to444(v422,v444);
    }
    else
    {
      conv422to444(src[1],u444);
      conv422to444(src[2],v444);
    }
  }

  strcat(outname, ".ppm");

  fprintf(stderr,"saving %s\n",outname);

  if ((outfile = open(outname,O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,0666))==-1)
  {
    sprintf(Error_Text,"Couldn't create %s\n",outname);
    Error(Error_Text);
  }

  optr = obfr;

  /* PPM header */
  sprintf(header,"P6\n%d %d\n255\n",horizontal_size,height);

  for (i=0; header[i]!=0; i++)
    putbyte(header[i]);

  /* matrix coefficients */
  crv = Inverse_Table_6_9[matrix_coefficients][0];
  cbu = Inverse_Table_6_9[matrix_coefficients][1];
  cgu = Inverse_Table_6_9[matrix_coefficients][2];
  cgv = Inverse_Table_6_9[matrix_coefficients][3];
  
  for (i=0; i<height; i++)
  {
    py = src[0] + offset + incr*i;
    pu = u444 + offset + incr*i;
    pv = v444 + offset + incr*i;

    for (j=0; j<horizontal_size; j++)
    {
      u = *pu++ - 128;
      v = *pv++ - 128;
      y = 76309 * (*py++ - 16); /* (255/219)*65536 */
      r = Clip[(y + crv*v + 32768)>>16];
      g = Clip[(y - cgu*u - cgv*v + 32768)>>16];
      b = Clip[(y + cbu*u + 32786)>>16];

      putbyte(r); putbyte(g); putbyte(b);
    }
  }

  if (optr!=obfr)
    write(outfile,obfr,optr-obfr);

  close(outfile);
}

unsigned char getbyte()
{
  unsigned char val;
  read(infile,&val,1);
  return val;
}

static void putbyte(c)
int c;
{
  *optr++ = c;

  if (optr == obfr+OBFRSIZE)
  {
    write(outfile,obfr,OBFRSIZE);
    optr = obfr;
  }
}

static void putword(w)
int w;
{
  putbyte(w); putbyte(w>>8);
}

/* horizontal 1:2 interpolation filter */
static void conv422to444(src,dst)
unsigned char *src,*dst;
{
  int i, i2, w, j, im3, im2, im1, ip1, ip2, ip3;

  w = Coded_Picture_Width>>1;

  if (MPEG2_Flag)
  {
    for (j=0; j<Coded_Picture_Height; j++)
    {
      for (i=0; i<w; i++)
      {
        i2 = i<<1;
        im2 = (i<2) ? 0 : i-2;
        im1 = (i<1) ? 0 : i-1;
        ip1 = (i<w-1) ? i+1 : w-1;
        ip2 = (i<w-2) ? i+2 : w-1;
        ip3 = (i<w-3) ? i+3 : w-1;

        /* FIR filter coefficients (*256): 21 0 -52 0 159 256 159 0 -52 0 21 */
        /* even samples (0 0 256 0 0) */
        dst[i2] = src[i];

        /* odd samples (21 -52 159 159 -52 21) */
        dst[i2+1] = Clip[(int)(21*(src[im2]+src[ip3])
                        -52*(src[im1]+src[ip2]) 
                       +159*(src[i]+src[ip1])+128)>>8];
      }
      src+= w;
      dst+= Coded_Picture_Width;
    }
  }
  else
  {
    for (j=0; j<Coded_Picture_Height; j++)
    {
      for (i=0; i<w; i++)
      {

        i2 = i<<1;
        im3 = (i<3) ? 0 : i-3;
        im2 = (i<2) ? 0 : i-2;
        im1 = (i<1) ? 0 : i-1;
        ip1 = (i<w-1) ? i+1 : w-1;
        ip2 = (i<w-2) ? i+2 : w-1;
        ip3 = (i<w-3) ? i+3 : w-1;

        /* FIR filter coefficients (*256): 5 -21 70 228 -37 11 */
        dst[i2] =   Clip[(int)(  5*src[im3]
                         -21*src[im2]
                         +70*src[im1]
                        +228*src[i]
                         -37*src[ip1]
                         +11*src[ip2]+128)>>8];

       dst[i2+1] = Clip[(int)(  5*src[ip3]
                         -21*src[ip2]
                         +70*src[ip1]
                        +228*src[i]
                         -37*src[im1]
                         +11*src[im2]+128)>>8];
      }
      src+= w;
      dst+= Coded_Picture_Width;
    }
  }
}

/* vertical 1:2 interpolation filter */
static void conv420to422(src,dst)
unsigned char *src,*dst;
{
  int w, h, i, j, j2;
  int jm6, jm5, jm4, jm3, jm2, jm1, jp1, jp2, jp3, jp4, jp5, jp6, jp7;

  w = Coded_Picture_Width>>1;
  h = Coded_Picture_Height>>1;

  if (progressive_frame)
  {
    /* intra frame */
    for (i=0; i<w; i++)
    {
      for (j=0; j<h; j++)
      {
        j2 = j<<1;
        jm3 = (j<3) ? 0 : j-3;
        jm2 = (j<2) ? 0 : j-2;
        jm1 = (j<1) ? 0 : j-1;
        jp1 = (j<h-1) ? j+1 : h-1;
        jp2 = (j<h-2) ? j+2 : h-1;
        jp3 = (j<h-3) ? j+3 : h-1;

        /* FIR filter coefficients (*256): 5 -21 70 228 -37 11 */
        /* New FIR filter coefficients (*256): 3 -16 67 227 -32 7 */
        dst[w*j2] =     Clip[(int)(  3*src[w*jm3]
                             -16*src[w*jm2]
                             +67*src[w*jm1]
                            +227*src[w*j]
                             -32*src[w*jp1]
                             +7*src[w*jp2]+128)>>8];

        dst[w*(j2+1)] = Clip[(int)(  3*src[w*jp3]
                             -16*src[w*jp2]
                             +67*src[w*jp1]
                            +227*src[w*j]
                             -32*src[w*jm1]
                             +7*src[w*jm2]+128)>>8];
      }
      src++;
      dst++;
    }
  }
  else
  {
    /* intra field */
    for (i=0; i<w; i++)
    {
      for (j=0; j<h; j+=2)
      {
        j2 = j<<1;

        /* top field */
        jm6 = (j<6) ? 0 : j-6;
        jm4 = (j<4) ? 0 : j-4;
        jm2 = (j<2) ? 0 : j-2;
        jp2 = (j<h-2) ? j+2 : h-2;
        jp4 = (j<h-4) ? j+4 : h-2;
        jp6 = (j<h-6) ? j+6 : h-2;

        /* Polyphase FIR filter coefficients (*256): 2 -10 35 242 -18 5 */
        /* New polyphase FIR filter coefficients (*256): 1 -7 30 248 -21 5 */
        dst[w*j2] = Clip[(int)(  1*src[w*jm6]
                         -7*src[w*jm4]
                         +30*src[w*jm2]
                        +248*src[w*j]
                         -21*src[w*jp2]
                          +5*src[w*jp4]+128)>>8];

        /* Polyphase FIR filter coefficients (*256): 11 -38 192 113 -30 8 */
        /* New polyphase FIR filter coefficients (*256):7 -35 194 110 -24 4 */
        dst[w*(j2+2)] = Clip[(int)( 7*src[w*jm4]
                             -35*src[w*jm2]
                            +194*src[w*j]
                            +110*src[w*jp2]
                             -24*src[w*jp4]
                              +4*src[w*jp6]+128)>>8];

        /* bottom field */
        jm5 = (j<5) ? 1 : j-5;
        jm3 = (j<3) ? 1 : j-3;
        jm1 = (j<1) ? 1 : j-1;
        jp1 = (j<h-1) ? j+1 : h-1;
        jp3 = (j<h-3) ? j+3 : h-1;
        jp5 = (j<h-5) ? j+5 : h-1;
        jp7 = (j<h-7) ? j+7 : h-1;

        /* Polyphase FIR filter coefficients (*256): 11 -38 192 113 -30 8 */
        /* New polyphase FIR filter coefficients (*256):7 -35 194 110 -24 4 */
        dst[w*(j2+1)] = Clip[(int)( 7*src[w*jp5]
                             -35*src[w*jp3]
                            +194*src[w*jp1]
                            +110*src[w*jm1]
                             -24*src[w*jm3]
                              +4*src[w*jm5]+128)>>8];

        dst[w*(j2+3)] = Clip[(int)(  1*src[w*jp7]
                             -7*src[w*jp5]
                             +30*src[w*jp3]
                            +248*src[w*jp1]
                             -21*src[w*jm1]
                              +5*src[w*jm3]+128)>>8];
      }
      src++;
      dst++;
    }
  }
}

void Error(text)
char *text;
{
  fprintf(stderr,text);
  exit(1);
}
